from unittest.mock import patch

import pytest
from orchestrator.types import SubscriptionLifecycle
from orchestrator.workflow import StepStatus
from pydantic_forms.exceptions import FormValidationError

from gso.products import Iptrunk
from gso.products.product_blocks.router import RouterRole
from test import USER_CONFIRM_EMPTY_FORM
from test.workflows import (
    assert_lso_interaction_success,
    assert_suspended,
    extract_state,
    resume_workflow,
    run_workflow,
)


@pytest.mark.parametrize("trunk_status", [SubscriptionLifecycle.PROVISIONING, SubscriptionLifecycle.ACTIVE])
@pytest.mark.parametrize("router_role", [RouterRole.P, RouterRole.PE])
@pytest.mark.workflow()
@patch("gso.services.lso_client._send_request")
@patch("gso.workflows.router.update_ibgp_mesh.librenms_client.LibreNMSClient.add_device")
@patch("gso.workflows.router.update_ibgp_mesh.librenms_client.LibreNMSClient.device_exists")
def test_update_ibgp_mesh_success(
    mock_librenms_device_exists,
    mock_librenms_add_device,
    mock_execute_playbook,
    router_role,
    trunk_status,
    iptrunk_subscription_factory,
    iptrunk_side_subscription_factory,
    router_subscription_factory,
    data_config_filename,
    faker,
):
    mock_librenms_device_exists.return_value = False
    side_a = iptrunk_side_subscription_factory(iptrunk_side_node=router_subscription_factory(router_role=router_role))
    side_b = iptrunk_side_subscription_factory()
    #  Add some extra devices to the network.
    router_subscription_factory(router_role=RouterRole.P)
    router_subscription_factory()

    ip_trunk = Iptrunk.from_subscription(
        iptrunk_subscription_factory(status=trunk_status, iptrunk_sides=[side_a, side_b])
    )
    callback_step_count = 5 if router_role == RouterRole.P else 14
    ibgp_mesh_input_form_data = {
        "subscription_id": ip_trunk.iptrunk.iptrunk_sides[0].iptrunk_side_node.owner_subscription_id
    }
    result, process_stat, step_log = run_workflow(
        "update_ibgp_mesh", [ibgp_mesh_input_form_data, {"tt_number": faker.tt_number()}]
    )

    for _ in range(callback_step_count):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    # Handle two consecutive user input steps
    for _ in range(2):
        assert_suspended(result)
        result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    state = extract_state(result)

    assert mock_execute_playbook.call_count == callback_step_count
    assert mock_librenms_add_device.call_count == 1
    assert result.status == StepStatus.COMPLETE
    assert state["subscription"]["router"]["router_access_via_ts"] is False


@pytest.mark.parametrize("trunk_status", [SubscriptionLifecycle.INITIAL, SubscriptionLifecycle.TERMINATED])
@pytest.mark.workflow()
def test_update_ibgp_mesh_failure(iptrunk_subscription_factory, data_config_filename, trunk_status):
    ip_trunk = Iptrunk.from_subscription(iptrunk_subscription_factory(status=trunk_status))
    ibgp_mesh_input_form_data = {
        "subscription_id": ip_trunk.iptrunk.iptrunk_sides[0].iptrunk_side_node.owner_subscription_id
    }

    exception_message = "Selected router does not terminate any available IP trunks."
    with pytest.raises(FormValidationError, match=exception_message):
        run_workflow("update_ibgp_mesh", [ibgp_mesh_input_form_data, {}])


@pytest.mark.workflow()
def test_update_ibgp_mesh_isolated_router(router_subscription_factory, data_config_filename):
    router_id = router_subscription_factory(router_role=RouterRole.P)

    exception_message = "Selected router does not terminate any available IP trunks."
    with pytest.raises(FormValidationError, match=exception_message):
        run_workflow("update_ibgp_mesh", [{"subscription_id": router_id}, {}])
